-- $LastChangedDate: 2008-11-12 18:30:45 -0500 (Wed, 12 Nov 2008) $
-- Actor classes.
-- See module.txt for documentation and pitfalls.
-- 
-- Although PlayerActor is a singleton, writing ":" is necessary,
-- despite it confusingly hinting of an instantiatable class.
-- For example, this will fail:
--   local name = PlayerActor.GetName()  -- wrong
--   local name = PlayerActor:GetName()  -- right
-- The reason is that the base Actor:Get*() methods require the "self" arg.
--
-- Derivative should completely override GetObject() and omit .object data member.
-- Reason is that multiple references to C++ Object complicates detaching from World.

---------------
-- functions --
---------------

function RandomMissileMalfunction()
	return math.random(10) <= 1
end

-----------------
-- Actor class --
-----------------

Actor = { }

-- ctor
function Actor:new( name, teamNum, object )
	if ( self == Actor ) then  -- being derived?
		self = lib.NewInstance( Actor )  -- no, need instance
	end

	self.name    = name
	self.teamNum = teamNum
	self.object  = object  -- ok if nil

  --Timer:Register( self, defs.ACTOR_TICK_FREQ )  -- wrong
	Timer:Register( self, defs.ACTOR_TICK_FREQ )  -- let derivative override base Tick()

	return self
end

-- In Lua, a destructor must be user-defined and explicitly called.
-- Order of destruction should be derivative-to-base.
function Actor:destroy()
	Timer:Unregister( self )
end

function Actor:GetObject()
	return self.object
end

function Actor:GetName()
	return self.name
end

function Actor:GetTeamNum()
	return self.teamNum
end

function Actor:IfShowInHud()
	return true
end

function Actor:Tick()
	-- NOP
end

function Actor:GetPosition()
	return self.object:GetPosition()
end

-------------------
-- AircraftActor --
-------------------

-- FYI:
-- AircraftCtor = { }
-- This is wrong if "AircraftCtor." is ever written.
-- Of course, because it's an empty table, it won't derive anything.

AircraftActor = lib.DerivedClass( Actor )

-- ctor.
function AircraftActor:new( name, teamNum, aircraftType, RadarClass )
	if ( self == AircraftActor ) then  -- either instantiate or use derivative's instance
		self = lib.NewInstance( AircraftActor )
	end

	Actor.new( self, name, teamNum, nil )

	self.aircraft     = sim.Aircraft:new( Vec3(0,0,0), aircraftType, sim.ModelCache.LOAD_COPY )
	self.aircraftType = aircraftType
	RadarClass = RadarClass or Radar
	self.radar        = RadarClass:new( self )

	return self
end

-- In Lua, a destructor must be user-defined and explicitly called.
-- Order of destruction should be derivative-to-base.
function AircraftActor:destroy()
	self.radar:destroy()  -- tell it to unreference any Actor/Object
	self.radar = nil

	Actor.destroy( self )
end

function AircraftActor:GetObject()
	return self.aircraft  -- self.object=nil
end

function AircraftActor:GetAircraft()
	return self.aircraft
end

function Actor:GetPosition()
	return self.aircraft:GetPosition()
end

function AircraftActor:Tick()
	self.radar:Tick()
end

-----------------------------
-- PlayerActor (singleton) --
-----------------------------

PlayerActor = lib.DerivedClass( AircraftActor )
PlayerActor.wingSweepFraction = 0
PlayerActor.wingSweepFractionIncForKey = 0.035
PlayerActor.wingSweepFractionIncForJoy = 0.010

-- Make/remake the player.
-- If aircraftType is nil, determine from cmd-line args.
function PlayerActor:new( aircraftType )
	-- In case of remake, first detach object before overwriting self.aircraft.
	if ( self.aircraft ) then
		sim.World:DetachObject( self.aircraft )
	end

	AircraftActor.new( self,
                       "PLAYER",
                       Game.TEAM_NUM_PLAYER,
                       aircraftType or self:GetAircraftType(),
                       PlayerRadar )

	self.replay         = Replay:new( self.aircraft )
	self.selectedTarget = nil  -- Actor
	self.savedState     = nil

	self:Init()

	return self
end

-- (private)
function PlayerActor:Init()
	sim.World:AttachObject( self.aircraft )
	self.aircraft:SetState( sim.Craft.eState_PARKED ) -- default
	sim.Control:SetCurrentCraft( self.aircraft )

	-- Enable coordinated-turning if this is a propeller aircraft.
	if ( sim.argv2["-coturn"] ) then
		self.aircraft:EnableCoordinatedTurn( true )
		Log( "Enabled coturn." )
	elseif ( sim.argv2["-no-coturn"] ) then
		self.aircraft:EnableCoordinatedTurn( false )
		Log( "Disabled coturn." )
	else
		self.aircraft:EnableCoordinatedTurn( self.aircraft:IfHasPropeller() )
	end
end

-- (private)
function PlayerActor:GetAircraftType()
	-- Parse cmdline args to select aircraft for player.
	aircraftType = sim.Aircraft.type_F14  -- default
	for k,v in pairs(sim.argv2) do
		if ( sim.Aircraft.typeArgMap[k] ) then
			aircraftType = sim.Aircraft.typeArgMap[k]
		end
	end
	return aircraftType
end

PlayerActor:new()  -- do it

-- Save player state.
function PlayerActor:Save()
	local physics = self.aircraft:GetPhysics()
	self.savedState = {
		aircraftState = self.aircraft:GetState(),
		matrix        = self.aircraft:GetMatrix(),
		throttle      = self.aircraft:GetThrottle(),
		speed         = physics:ComputeSpeed(),
		landingGear   = self.aircraft:IfLandingGear(),
	}
end

-- Reset player.
function PlayerActor:Reset()
	-- Restore player from saved state.
	if ( self.savedState ) then
		local physics = self.aircraft:GetPhysics()
		BeginPlacement( self.aircraft )  -- disable collisions
		self.aircraft:Reset()
		self.aircraft:SetState(       self.savedState.aircraftState )
		self.aircraft:SetMatrix(      self.savedState.matrix )
		self.aircraft:SetThrottle(    self.savedState.throttle )
		self.aircraft:SetLandingGear( self.savedState.landingGear )
		physics:SetSpeed(             self.savedState.speed )
		EndPlacement( self.aircraft )
	end

	-- Reset views.
	ChaseView:Init( sim.argv2["-ossim"] )
	sim.SuperView:Reset()

	-- Reset lights.
	Brightness:Reset()

	-- SUBTLE: Although the Replay object is reassigned and no longer referenced,
	-- Lua delays garbage-collecting it.  Replay::Stop() is called to stop
	-- its timer-ticks so that it becomes inactive (zombie) else it would
	-- content with the new Replay object!
	self.replay:Stop()  -- disable its timer-ticks (turn into zombie)
	self.replay = Replay:new( self.aircraft )  -- remake
end

Events:RegisterReset( function () PlayerActor:Reset() end )  -- need anon func

-- Get Replay object.
function PlayerActor:GetReplay()
	return self.replay
end

-- Get target selected by player else nil.
function PlayerActor:GetSelectedTarget()
	return self.selectedTarget
end

-- Let player select a target (show in HUD).
function PlayerActor:SelectTarget( actor, radarLock )
	-- Occasionally actor=nil (why?).
	if ( actor ) then
		-- Lua
		self.selectedTarget = actor

		-- C++
		sim.GameHud:SelectTarget( actor:GetObject(), radarLock or false )
	end
end

-- Switch to next target.
function PlayerActor:CycleTargets()
	-- Is a target selected?
	-- If not, choose the first actor.
	local nextSelectedTarget = nil
	if ( self.selectedTarget ) then
		nextSelectedTarget = Game:GetNextActor( self.selectedTarget, Game.TEAM_NUM_ENEMY )
	else
		nextSelectedTarget = Game:GetActors()[1]
	end

	-- Update HUD if the next target was selected.
	if ( nextSelectedTarget ) then
		self:SelectTarget( nextSelectedTarget )  -- assigns self.selectedTarge var
	end
end

-- Incrementally adjust wing-sweep.
-- +1:forward or -1:backward
-- input = "key" or "joy"
function PlayerActor:IncWingSweep( dir, input )
	-- Inc or dec.
	-- But at a slow rate for joystick which is polled faster than keyboard.
	local inc = self.wingSweepFractionIncForKey
	if ( input == "joy" ) then
		inc = self.wingSweepFractionIncForJoy
	end

	-- Compute fraction of wing sweep (0.0,...,1.0).
	local wingSweepFraction = self.wingSweepFraction + dir * inc

	-- Apply wing-sweep (unless outside range).
	if ( wingSweepFraction >= 0 and wingSweepFraction <= 1 ) then
		self.wingSweepFraction = wingSweepFraction -- commit
		self.aircraft:SweepWings( wingSweepFraction )
	end
end

-- Fire selected weapon.
function PlayerActor:FireWeapon()
	self:FireMissile()
end

-- Fire guns.
function PlayerActor:FireGun()
	print( "fire gun TODO" )
end

-- Fire missile.
function PlayerActor:FireMissile()
	-- Does radar have lock?
	local target = self.radar:GetTarget()
	if ( not target ) then
		Hud:AddStatusText( defs.HUD_STATUS_NO_RADAR_LOCK )
	else
		Hud:RemoveStatusText( defs.HUD_STATUS_NO_RADAR_LOCK )

		-- Create missile (as an Actor).
		local launcher = PlayerActor
		local launcherSpeed = self.aircraft:GetPhysics():ComputeSpeed()
		local malfunction = RandomMissileMalfunction()
		local particleRadiusGeo = 8.0  -- smaller trail for player
		local missile = MissileActor:new( target, launcher, launcherSpeed,
                                          particleRadiusGeo, malfunction )

		if ( malfunction ) then
			Hud:AddStatusText( defs.HUD_STATUS_MISSILE_MALFUNCTION )
		else
			Hud:RemoveStatusText( defs.HUD_STATUS_MISSILE_MALFUNCTION )
		end

		-- Contain missile.
		Game:AttachActor( missile )

		-- Play missile sound.
		sim.Sound:Play( "missile.wav.gz", sim.Sound.PLAY_ONCE, 1.0 )
	end
end

-- Abbrev.
function GetPlayerAircraft()
	return PlayerActor:GetAircraft()
end

-------------------
-- MissileActor --
-------------------

MissileActor = lib.DerivedClass( Actor )

-- ctor.
function MissileActor:new( target, launcher, launcherSpeed, particleRadiusGeo, malfunction )
	if ( self == MissileActor ) then  -- either instantiate or use derivative's instance
		self = lib.NewInstance( MissileActor )
	end

	Actor.new( self, "missile", launcher:GetTeamNum(), nil )

	self.target  = target
	self.missile = sim.Missile:new( target:GetObject(), launcher:GetObject(),
                                    launcherSpeed, particleRadiusGeo, malfunction )
	return self
end

function MissileActor:destroy()
	self.missile:destroy()  -- tell it to unreference any Actor/Object

	Actor.destroy( self )
end

-- N/A (C++ doesn't expose 3D Object).
function MissileActor:GetObject()
	return nil
end

-- Don't show in HUD.
function MissileActor:IfShowInHud()
	return false
end

function MissileActor:Tick()
	-- NOP
end

-- Orient missile.
function MissileActor:Rotate( axis, rad )
	self.missile:Rotate( axis, rad )
end

--------------
-- SamActor --
--------------

SamActor = lib.DerivedClass( Actor )

-- ctor.
function SamActor:new( name, teamNum, object )
	if ( self == SamActor ) then  -- either instantiate or use derivative's instance
		self = lib.NewInstance( SamActor )
	end

	Actor.new( self, name, teamNum, object )

	self.target  = PlayerActor -- ### TODO/HACK/FIXME ###
	self.missile = nil
	self.detection = false

	return self
end

function SamActor:destroy()
	-- If missile was created.
	if ( self.missile ) then
		self.missile:destroy()  -- tell it to unreference any Actor/Object
	end

	Actor.destroy( self )
end

-- private (Mission:Reset() calls this)
function SamActor:Reset()
    self:GetObject():SetCollision( sim.COLLISION_NONE )
	self.detection = false
end

-- private
function SamActor:LaunchMissile()
	-- Create MissileActor.
	local target = PlayerActor
	local launcher = self
	local speed = sim.Physics.KPH2Speed( 600.0 )  -- slow to make game easy
	--speed = PlayerActor:GetAircraft():GetPhysics():ComputeSpeed()
	local particleRadiusGeo = M2W( 20.0 )  -- larger
	local malfunction = RandomMissileMalfunction()
	self.missile = MissileActor:new( target, launcher, speed, particleRadiusGeo, malfunction )
	
	-- Orient missile to shoot upward.
	self.missile:Rotate( AXIS_PITCH, Deg2Rad( 30.0 ) )
	self.missile:Rotate( AXIS_YAW, Deg2Rad( 20.0 ) )
	
	-- Attach MissileActor.
	Game:AttachActor( self.missile )
end

-- Tick to pulse animation.
function SamActor:Tick()
	-- Launch one missile every time player transitions into the range of SAM.
    if ( not self:GetObject():IfCollision() ) then
		local distance = Sim2Meters( Distance3( PlayerActor:GetPosition(), self.object:GetPosition() ) )
		if ( distance < defs.SAM_RADAR_RANGE_METERS ) then
			-- Within range but is this a transition?
			if ( not self.detection ) then
				self.detection = true -- remember
				self.launchedMissile = true
				self:LaunchMissile()
				Hud:AddStatusText( defs.HUD_STATUS_MISSILE_WARNING )
				sim.Sound:Play( "beep2.wav.gz", sim.Sound.PLAY_ONCE, 1.0 )
			end
		else
			-- Outside range.
			self.detection = false -- remember
		end
	end
end
