/*****************************************************************************
 * $LastChangedDate: 2008-11-05 23:44:26 -0500 (Wed, 05 Nov 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Special-effects container.
 *//*
 * LEGAL:	COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#define FX_FX_CC 1
#include "base/module.hh"
using namespace base;
#include "object/module.hh"
using namespace object;
#include "world/module.hh"
using namespace world;
#include "fx/module.hh"
#include "fx/fx.hh"
#include "fx/defs.hh"
#include "fx/fx_partsys.hh"
#include "fx/fx_partsys_explosion.hh"
#include "fx/fx_partsys_trail.hh"

namespace fx {

////////////////////////////////////////////////////////////////////////////////
/////////////////////////////////  FX : core  //////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

FX* FX::msInstance;

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
FX::FX( void )
:	mTimerFunctor(new FX::TimerFunctor(*this)),
	mObjects(),
	mZombieObjects(),
	mParticleSystems(),
	mZombieParticleSystems()
{
	Timer::GetInstance().Register( mTimerFunctor.PTR(), fx::defs::FX_TICK_FREQ );
}

FX::~FX()
{
	// Must unregister!  Else functor would access a destroyed object.
	Timer::GetInstance().Unregister( mTimerFunctor.PTR() );
}

/*****************************************************************************
 * Stop all special-effects.
 *****************************************************************************/
void
FX::Reset( void )
{
	// Detach Objects.
	Detach( mObjects );
	Detach( mZombieObjects );

	// Detach ParticleSystems.
	Detach( mParticleSystems );
	Detach( mZombieParticleSystems );
}

/*****************************************************************************
 * Periodically destroy zombie special-effects.
 *****************************************************************************/
void
FX::Tick( const Milliseconds millisecElapsed )
{
	Detach( mZombieObjects );
	Detach( mZombieParticleSystems );
}

// Reflect.
void
FX::TimerFunctor::operator()( const Milliseconds millisecElapsed )
{
	mFX.Tick( millisecElapsed );
}

/*****************************************************************************
 * Detach special-effects (Objects).
 *****************************************************************************/
void
FX::Detach( Objects& objects )
{
	// Detach all special-effects out of scene graph.
	for ( Objects::iterator iter = objects.begin(); iter != objects.end(); ++iter )
	{
		CHECK_TYPESIG( *iter, TYPESIG_OBJECT );  // *iter = shptr<Object>
		gWorld.Detach( *iter );
	}

	// Erase container.
	objects.clear();
}

/*****************************************************************************
 * Detach special-effects (ParticleSystemFX).
 *****************************************************************************/
void
FX::Detach( ParticleSystems& particleSystems )
{
#if DEBUG_PARTSYS
	CDEBUG << "FX::DetachParticleSystems() cnt=" << particleSystems.size() << endl;
#endif

	// Erase container.  ParticleSystems aren't 3D Objects and outside scene graph.
	particleSystems.clear();
}

/*****************************************************************************
 * Turn Object into a zombie to be deleted later.
 *****************************************************************************/
void
FX::AddZombie( shptr<Object> zombie )
{
CHECK_TYPESIG(zombie,TYPESIG_OBJECT);

	// Add to zombie list.  Re-adding would be a benign error because of shptr.
	mZombieObjects.push_back( zombie );

	// Remove from active list.
	for ( Objects::iterator iter = mObjects.begin();
          iter != mObjects.end();
          ++iter )
	{
		CHECK_TYPESIG( *iter, TYPESIG_OBJECT );  // *iter = shptr<Object>
		if ( (*iter).PTR() == zombie.PTR() )  // compare address
		{
			gWorld.Detach( *iter );
			mObjects.erase( iter );  // ok since going to return
			return;
		}
	}
	ASSERT( false );  // oops, it should be in active list
}

/*****************************************************************************
 * Turn ParticleSystemFX into a zombie to be deleted later.
 *****************************************************************************/
void
FX::AddZombie( shptr<ParticleSystemFX> zombie )
{
CHECK_TYPESIG(zombie,TYPESIG_PARTICLE_SYSTEM);

	// Add to zombie list.  Re-adding would be a benign error because of shptr.
	mZombieParticleSystems.push_back( zombie );

	// Remove from active list.
	for ( ParticleSystems::iterator iter = mParticleSystems.begin();
          iter != mParticleSystems.end();
          ++iter )
	{
		CHECK_TYPESIG(*iter,TYPESIG_PARTICLE_SYSTEM);  // *iter = shptr<ParticleSystemFX>
		if ( (*iter).PTR() == zombie.PTR() )  // compare address
		{
			mParticleSystems.erase( iter );  // ok since going to return
			return;
		}
	}
	ASSERT( false );  // oops, it should be in active list
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////  FX : make special-effects  ///////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * Create an explosion.
 *****************************************************************************/
void
FX::MakeExplosion( const WorldVertex& position,
                   const fp radius,
                   const Milliseconds lifetime )
{
	// Create a ParticleSystem.  Add to FX container.
	mParticleSystems.push_back( new ExplosionParticleSystemFX( position, radius, lifetime ) );
}

/*****************************************************************************
 * Create smoke trails.
 *****************************************************************************/
void
FX::MakeSmokeTrail( shptr<Object> trailMaker,
                   const fp radius,
                   const Milliseconds lifetime )
{
	// Create a ParticleSystem.  Add to FX container.
	mParticleSystems.push_back( new TrailParticleSystemFX( trailMaker, radius, lifetime ) );
}

} // namespace fx
