/*****************************************************************************
 * $LastChangedDate: 2008-11-12 00:53:55 -0500 (Wed, 12 Nov 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Trail particle-system (private implementation of FX class).
 *//*
 * LEGAL:	COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

// TODO: Listen for destroy-Object event then destroy partsys.

#define COMPILE_FADE_TRAIL_PARTICLE 0

#define FX_FX_PARTSYS_TRAIL_CC 1
#include "base/module.hh"
#include "base/random.hh"
using namespace base;
#include "object/module.hh"
using namespace object;
#include "fx/module.hh"
#include "fx/texture.hh"
#include "fx/fx_partsys_trail.hh"
#include "fx/defs.hh"

namespace fx {

INTERN const Milliseconds   TRAIL_PARTICLE_LIFETIME( 25*1000 );
INTERN const uint           TRAIL_PARTICLE_LIMIT = 3000;
INTERN const uint           TRAIL_PARTICLES_PER_SECOND( 60 );
INTERN const RGBA           TRAIL_COLOR_RED(  0xff, 0x77, 0x11, 0xa0 );
INTERN const RGBA           TRAIL_COLOR_GRAY( 0xc0, 0xc0, 0xc0, 0xa0 );
#define                     TRAIL_FIRE_TIME  250

////////////////////////////////////////////////////////////////////////////////
//////////////////////////////  TrailParticle  /////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 *****************************************************************************/
TrailParticle::TrailParticle( const Sprite::Args& spriteArgs )
:	Particle(spriteArgs,TRAIL_PARTICLE_LIFETIME)
{
	SetColor0( TRAIL_COLOR_RED );
}

TrailParticle::~TrailParticle()
{
	// NOP
}

/*****************************************************************************
 * Animate particle.
 *****************************************************************************/
void
TrailParticle::Tick( const Milliseconds millisecElapsed )
{
CHECK_TYPESIG(this,TYPESIG_PARTICLE);

	// Fade-out particle.
    // 1.0 : begin, 0.0 : end
    const Milliseconds age = GetAge( millisecElapsed, TRAIL_PARTICLE_LIFETIME, mTimeOfDeath );
   	RGBA color = GetColor0();

    // Compute color.
    // Particles initiailly burn red then change to gray and fade away.
    if ( age < Milliseconds(TRAIL_FIRE_TIME) )
    {
        color = Interpolate( TRAIL_COLOR_RED,
                             TRAIL_COLOR_GRAY,
                             fp(age) / TRAIL_FIRE_TIME );
    }
    else
    {
#if COMPILE_FADE_TRAIL_PARTICLE
        const fp delta = fp(mTimeOfDeath) - fp(millisecElapsed);
        const fp frac = delta / fp(TRAIL_PARTICLE_LIFETIME);
        color[RR] = \
        color[GG] = \
        color[BB] = 0xc0;
        color[AA] = ClampFF( int(0xff * frac) );
#else
        color[RR] = \
        color[GG] = \
        color[BB] = 0xc0;
        color[AA] = 0xff;
#endif
    }
    SetColor0( color );
}

////////////////////////////////////////////////////////////////////////////////
///////////////////////////  TrailParticleSystemFX  ////////////////////////////
////////////////////////////////////////////////////////////////////////////////

/*****************************************************************************
 * ctor/dtor.
 * @param	trailMaker
 *			Emitter of particles (missile, damaged aircraft, etc).
 *****************************************************************************/
TrailParticleSystemFX::TrailParticleSystemFX( shptr<Object> trailMaker,
                                              const fp radius,
                                              const Milliseconds lifetime )
	// For state-sorting, place all particles under one group node sharing the same texture.
:	ParticleSystemFX(lifetime, ParticleSystem::DEFAULT_STATE_BITS | StateSorter::AllocateStateBitsId()),
    mTrailMaker(trailMaker),
    mTrailMakerPrevPos(trailMaker->GetPosition()),
	mRadius(radius),
	mTimeOfDeath(base::Timer::GetInstance().GetElapsedTime() + TRAIL_PARTICLE_LIFETIME),
	mTimeOfPrevTick(base::Timer::GetInstance().GetElapsedTime()),
	// For state-sorting, create stateset to be shared by all particles.
	mStateSet(Texture::NewTextureAsStateSet("textures/smoke.png"))
{
ASSERT( radius > 0 );

	// NOP.  Tick() does it.
}

TrailParticleSystemFX::~TrailParticleSystemFX()
{
	// NOP
}

/*****************************************************************************
 * Every tick, decide:
 * - whether to expire entire particle system
 *   (base ParticleSystemFX decides this)
 * - whether to create a new Particle
 *****************************************************************************/
void
TrailParticleSystemFX::Tick( const Milliseconds millisecElapsed )
{
CHECK_TYPESIG(this,TYPESIG_PARTICLE_SYSTEM);

	// Call parent's method.  It may expire this.
	Parent::Tick( millisecElapsed );

	// Stop emitting new particiles if partsys was expired earlier,
    // or missile was destroyed (but keep particles for latter case).
	if ( (not IfExpired()) and (not mTrailMaker->IfZombie()) )
	{
		// Kill the oldest particles to make way for new ones.
		// Let hot-fire particles replace old-smoke particles.
		const int excess = GetParticleCnt() - TRAIL_PARTICLE_LIMIT;
		if ( excess > 0 )
		{
			DetachOldestParticles( excess );
		}

		// From time between now and previous tick,
		// compute how many Particles to create.
		const Milliseconds timeDelta = millisecElapsed - mTimeOfPrevTick;
		ASSERT( timeDelta >= Milliseconds(0) );
		uint particleCnt = Max( uint(1),
                                uint( fp(timeDelta) / 1000.0f * TRAIL_PARTICLES_PER_SECOND ) );
		MakeParticles( particleCnt );
	}

	// Update time-stamp.
	mTimeOfPrevTick = millisecElapsed;
    mTrailMakerPrevPos = mTrailMaker->GetPosition();
}

/*****************************************************************************
 * Make Particles.
 *****************************************************************************/
void
TrailParticleSystemFX::MakeParticles( const uint cnt )
{
CHECK_TYPESIG(this,TYPESIG_PARTICLE_SYSTEM);
ASSERT( cnt < fx::defs::MAX_PARTICLES_ONE_SYSTEM );

	// Create/attach new particles.
    // Interpolate particles across current step of moving trail maker.
    WorldVertex pos = mTrailMakerPrevPos;
    WorldVertex posInc = (mTrailMaker->GetPosition() - mTrailMakerPrevPos) / fp(cnt);
    for ( uint i = 0; i < cnt; ++i )
    {
        const fp radius = mRadius * (1.0f + gRandom.random_f(0.5f));
        Sprite::Args spriteArgs( pos,
                                 radius,
                                 mStateSet,
                                 RGBA() );  // color passed isn't important
    	AttachParticle( new TrailParticle( spriteArgs ) );  // implied stateBits from ctor arg
        pos += posInc;  // linear interpolation
    }
}

} // namespace fx
