/*****************************************************************************
 * $LastChangedDate: 2008-11-12 13:49:48 -0500 (Wed, 12 Nov 2008) $
 * @file
 * @author	Jim E. Brooks  http://www.palomino3d.org
 * @brief	Missiles.
 *//*
 * LEGAL:   COPYRIGHT (C) 2008 JIM E. BROOKS
 *          THIS SOURCE CODE IS RELEASED UNDER THE TERMS
 *          OF THE GNU GENERAL PUBLIC LICENSE VERSION 2 (GPL 2).
 *****************************************************************************/

#ifndef GAME_MISSILE_HH
#define GAME_MISSILE_HH 1

#include "physics/module.hh"
using namespace physics;

namespace game {

////////////////////////////////////////////////////////////////////////////////
/// @brief Missile's underlying 3D object.
///
/// The purposes of MissileDyna:
/// - To define GetName() as "Missile" which is needed for crash determination.
/// - The Missile class ctor has smart-pointer problems if derived from Dyna.
///   So, Missile has a MissileDyna (instead of being a Dyna).
///
class MissileDyna : public Dyna
{
public:
                            MissileDyna( shptr<Graph> graph, const WorldVertex& pos )
                            : Dyna(graph,pos) { }
    virtual const string	GetName( void ) { return string("Missile"); }
};

////////////////////////////////////////////////////////////////////////////////
/// @brief Missiles.
///
/// Lua:
/// ----
/// Lua determines which Actor is the target.
/// Lua interfaces with this C++ class in terms of the Dyna class.
///
/// C++:
/// ----
/// Missile has a MissileDyna (3D Object) loaded with 3D model of a missile.
/// Missile inherits its matrix and speed from its launcher.
/// Every tick, a guidance formula rotates the Dyna's matrix
/// to guide the missile at the target.
/// If the target is hit, movement stops.
/// Missile does its own distance-based collision-detection.
/// If the target is destroyed by another action,
/// missile will fly straight indefinitely.
///
class Missile : public Shared  // there is no C++ Actor class
{
PREVENT_COPYING(Missile)
public:
                    	Missile( shptr<Object> target,
                                 Object& launcher,
                                 const Speed launcherSpeed,
                                 const Meter particleRadiusGeo,
                                 const bool malfunction );
                        ~Missile();
    void                Rotate( uint axis, Radian rad );

private:
    CLASS_METHOD shptr<MissileDyna> MakeMissile( Object& launcher );
    void                Tick( const Milliseconds millisecElapsed );
    WorldVertex         Guide( const fp step1 );
    void                Maneuver( Matrix& matrix,  /*IN/OUT*/
                                  const uint rotationAxis,
                                  const uint nextRotationAxis,
                                  Radian maneuverAngle, 
                                  const WorldVertex targetPos );
private:
	/// @brief Timer-tick.
    CLASS_TICK_FUNCTOR( Missile, mMissile )

private:
    // Graphical:
    shptr<Object>       mTarget;
    shptr<MissileDyna>  mMissile;           ///< Missile has a Dyna (Missile isn't a Dyna)

    // Physical:
    SpeedKPH            mSpeedKPH;          ///< scalar speed

    // Guidance:
    WorldVertex         mDirection;         ///< forward step
    uint                mRotationAxis;      ///< AXIS_ROLL or AXIS_PITCH
	Radian              mManeuverAngleRoll;
	Radian              mManeuverAnglePitch;
	fp                  mPredictedManeuverAngleSign;	///< try rotating in same direction as previously

    // Animation:
	shptr<TickFunctor>	mTickFunctor;
    Milliseconds        mPrevTick;
    const WorldVertex   mLauncherPosition;
    bool                mCollidable;        ///< temporarily indestructible while launching
	bool                mHit;               ///< missile is zombified after hit
    const fp            mHitDistance;       ///< scalar in world space
    const bool          mMalfunction;       ///< missiles don't always work
public:
    DECL_TYPESIG(TYPESIG_MISSILE)
};

} // namespace game

#endif // GAME_MISSILE_HH
